(*
 /************************************************************************************/
 /*                                                                                  */
 /*           This program has been converted from C to Delphi Pascal                */
 /*                                                                                  */
 /*                       by Paul C. L. Willmott, VP9MU                              */
 /*                                                                                  */
 /*                              July 11, 2003                                       */
 /*                                                                                  */
 /************************************************************************************/

 /************************************************************************************/
 /*                                                                                  */
 /*        This program has been adapted by VK5HI to process FEC Files               */
 /*                                                                                  */
 /*              as an evaluative tool using command line execution.                 */
 /*                                                                                  */
 /*                     Only the main() has been modified                            */
 /*                                                                                  */
 /*            The main() also includes James's sync_detection code                  */
 /*                                                                                  */
 /*        James's original main() is at end of program as main_Original()           */
 /*                                                                                  */
 /*                                                                                  */
 /************************************************************************************/

 /************************************************************************************/
 /*                                                                                  */
 /*     Reference Decoder for proposed AMSAT P3 Satellite Telemetry FEC Format       */
 /*                                                                                  */
 /*                              Version 1.0                                         */
 /*                                                                                  */
 /*                Last modified 2003 Jun 22 [Sun] 1035 utc                          */
 /*                                                                                  */
 /*                      (C)2003 J R Miller G3RUH                                    */
 /*                                                                                  */
 /*  This program is based on material largely developed by Phil Karn KA9Q over the  */
 /*  period 1997-2002.   Much of it can be found in libraries on his www site at     */
 /*  http://www.ka9q.net/ao40/  and includes notices which are reproduced herein.    */
 /*                                                                                  */
 /************************************************************************************/
*)

unit FEC ;

interface

implementation

// Defines for environment

(*
#include <stdio.h>     /* for: fopen  fread  fwrite  fclose  fprintf   */
#include <stdlib.h>    /* for: calloc  free */
#include <string.h>    /* for: memset  memove  memcpy */
#include <io.h>		     /* for filelength */
*)

const

       Verbose = 1 ;   // Permit some diagnostics; use 0 for none

//     Defines for RS Decoder(s)

(*
function minm(a,b:integer):integer ; overload ; //       ((a) < (b) ? (a) : (b))
function minm(a,b:double):double ; overload ;   //       ((a) < (b) ? (a) : (b))


use Min internal function
*)

NN     = 255 ;
KK     = 223 ;
NROOTS =  32 ;   // NN-KK
FCR    = 112 ;
PRIM   =  11 ;
IPRIM  = 116 ;
A0     =  NN ;

BLOCKSIZE = 256 ; // Data bytes per frame
RSBLOCKS  =   2 ; // Number of RS decoder blocks
RSPAD     =  95 ; // Unused bytes in block  (KK-BLOCKSIZE/RSBLOCKS)

// Defines for Viterbi Decoder for r=1/2 k=7  (to CCSDS convention)

K      = 7   ; // Constraint length
N      = 2   ; // Number of symbols per data bit
CPOLYA = $4f ; // First  convolutional encoder polynomial
CPOLYB = $6d ; // Second convolutional encoder polynomial

// Number of symbols in an FEC block that are
// passed to the Viterbi decoder  (320*8 + 6)

NBITS = (BLOCKSIZE+NROOTS*RSBLOCKS)*8+K-1 ;


// Defines for Re-encoder

SYNC_POLY = $48 ; // Sync vector PN polynomial

// Declarations for Reference encoder

type

   TEncodedSymbols : packed array [0..SYMPBLOCK-1] of byte ;  // Encoded symbols
   TSymbols        : packed array [0..5131] of byte ;         // De-interleave Output Data
   TVicDecData     : packed array [0..319] of byte ;          // Viterbi decoder output data
   TRSDecData      : packed array [0..BLOCKSIZE-1] of byte ;  // RS decoder data output (to user)*/

var

   Nbytes   : integer ;                                   // Byte counter for encode_data()
   Bindex   : integer ;                                   // Byte counter for interleaver
   Conv_sr  : byte ;                                      // Convolutional encoder shift register state
   RS_block : array [0..RSBLOCKS-1,0..NROOTS-1] of byte ; // RS parity blocks
   reencode : TReEncodedSymbols ;                         // Re-encoded symbols

const

   RS_poly : array [0..15] of byte = (249, 59, 66, 4, 43, 126, 251, 97, 30, 3, 213, 50, 66, 170, 5, 24) ;

(*   Amsat P3 FEC Encoder/decoder system.   Look-up tables
     created by Phil Karn KA9Q and James Miller G3RUH
     Last modified 2003 Jun 20  *)

// 8-bit parity table

Partab : array[0..255] of byte = (
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  1, 0, 0, 1, 0, 1, 1, 0, 0, 1, 1, 0, 1, 0, 0, 1,
  0, 1, 1, 0, 1, 0, 0, 1, 1, 0, 0, 1, 0, 1, 1, 0
} ;

(* Tables for Viterbi r=1/2 k=7 decoder to CCSDS standard */
/* ------------------------------------------------------ */

/* Metric table, [sent sym][rx symbol] */
/* This metric table is for an 8-bit ADC, which is total overkill!
   Simplify later.  128-i and i-128 would probably do!  jrm *)

mettab : array[0..1] of array[0..255] of integer =
((
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   19,   19,   19,   19,   19,   19,   19,   19,   19,
  19,   19,   18,   18,   18,   18,   18,   18,   17,   17,   17,   16,   16,   16,   15,   15,
  14,   14,   13,   13,   12,   11,   10,   10,    9,    8,    7,    6,    5,    3,    2,    1,
  -1,   -2,   -4,   -5,   -7,   -9,  -11,  -13,  -15,  -17,  -19,  -21,  -23,  -25,  -28,  -30,
 -32,  -35,  -37,  -40,  -42,  -45,  -47,  -50,  -52,  -55,  -58,  -60,  -63,  -66,  -68,  -71,
 -74,  -77,  -79,  -82,  -85,  -88,  -90,  -93,  -96,  -99, -102, -104, -107, -110, -113, -116,
-119, -121, -124, -127, -130, -133, -136, -138, -141, -144, -147, -150, -153, -155, -158, -161,
-164, -167, -170, -172, -175, -178, -181, -184, -187, -190, -192, -195, -198, -201, -204, -207,
-210, -212, -215, -218, -221, -224, -227, -229, -232, -235, -238, -241, -244, -247, -249, -252,
-255, -258, -261, -264, -267, -269, -272, -275, -278, -281, -284, -286, -289, -292, -295, -298,
-301, -304, -306, -309, -312, -315, -318, -320, -324, -326, -329, -332, -335, -337, -341, -372),
(
-372, -341, -338, -335, -332, -329, -326, -324, -321, -318, -315, -312, -309, -306, -304, -301,
-298, -295, -292, -289, -286, -284, -281, -278, -275, -272, -269, -267, -264, -261, -258, -255,
-252, -249, -247, -244, -241, -238, -235, -232, -229, -227, -224, -221, -218, -215, -212, -210,
-207, -204, -201, -198, -195, -192, -190, -187, -184, -181, -178, -175, -172, -170, -167, -164,
-161, -158, -155, -153, -150, -147, -144, -141, -138, -136, -133, -130, -127, -124, -121, -119,
-116, -113, -110, -107, -104, -102,  -99,  -96,  -93,  -90,  -88,  -85,  -82,  -79,  -77,  -74,
 -71,  -68,  -66,  -63,  -60,  -58,  -55,  -52,  -50,  -47,  -45,  -42,  -40,  -37,  -35,  -32,
 -30,  -28,  -25,  -23,  -21,  -19,  -17,  -15,  -13,  -11,   -9,   -7,   -5,   -4,   -2,   -1,
   1,    2,    3,    5,    6,    7,    8,    9,   10,   10,   11,   12,   13,   13,   14,   14,
  15,   15,   16,   16,   16,   17,   17,   17,   18,   18,   18,   18,   18,   18,   19,   19,
  19,   19,   19,   19,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20
});


(* Table of symbol pair emitted for each state of the convolutional
 encoder's shift register *)

Syms : array[0..127] of integer = (
  1, 2, 3, 0, 2, 1, 0, 3, 2, 1, 0, 3, 1, 2, 3, 0,
  1, 2, 3, 0, 2, 1, 0, 3, 2, 1, 0, 3, 1, 2, 3, 0,
  0, 3, 2, 1, 3, 0, 1, 2, 3, 0, 1, 2, 0, 3, 2, 1,
  0, 3, 2, 1, 3, 0, 1, 2, 3, 0, 1, 2, 0, 3, 2, 1,
  2, 1, 0, 3, 1, 2, 3, 0, 1, 2, 3, 0, 2, 1, 0, 3,
  2, 1, 0, 3, 1, 2, 3, 0, 1, 2, 3, 0, 2, 1, 0, 3,
  3, 0, 1, 2, 0, 3, 2, 1, 0, 3, 2, 1, 3, 0, 1, 2,
  3, 0, 1, 2, 0, 3, 2, 1, 0, 3, 2, 1, 3, 0, 1, 2
);


// Scramble byte table

Scrambler : array[0..255] of byte =
  (
   $ff, $48, $0e, $c0, $9a, $0d, $70, $bc, $8e, $2c, $93, $ad, $a7, $b7, $46, $ce,
   $5a, $97, $7d, $cc, $32, $a2, $bf, $3e, $0a, $10, $f1, $88, $94, $cd, $ea, $b1,
   $fe, $90, $1d, $81, $34, $1a, $e1, $79, $1c, $59, $27, $5b, $4f, $6e, $8d, $9c,
   $b5, $2e, $fb, $98, $65, $45, $7e, $7c, $14, $21, $e3, $11, $29, $9b, $d5, $63,
   $fd, $20, $3b, $02, $68, $35, $c2, $f2, $38, $b2, $4e, $b6, $9e, $dd, $1b, $39,
   $6a, $5d, $f7, $30, $ca, $8a, $fc, $f8, $28, $43, $c6, $22, $53, $37, $aa, $c7,
   $fa, $40, $76, $04, $d0, $6b, $85, $e4, $71, $64, $9d, $6d, $3d, $ba, $36, $72,
   $d4, $bb, $ee, $61, $95, $15, $f9, $f0, $50, $87, $8c, $44, $a6, $6f, $55, $8f,
   $f4, $80, $ec, $09, $a0, $d7, $0b, $c8, $e2, $c9, $3a, $da, $7b, $74, $6c, $e5,
   $a9, $77, $dc, $c3, $2a, $2b, $f3, $e0, $a1, $0f, $18, $89, $4c, $de, $ab, $1f,
   $e9, $01, $d8, $13, $41, $ae, $17, $91, $c5, $92, $75, $b4, $f6, $e8, $d9, $cb,
   $52, $ef, $b9, $86, $54, $57, $e7, $c1, $42, $1e, $31, $12, $99, $bd, $56, $3f,
   $d2, $03, $b0, $26, $83, $5c, $2f, $23, $8b, $24, $eb, $69, $ed, $d1, $b3, $96,
   $a5, $df, $73, $0c, $a8, $af, $cf, $82, $84, $3c, $62, $25, $33, $7a, $ac, $7f,
   $a4, $07, $60, $4d, $06, $b8, $5e, $47, $16, $49, $d6, $d3, $db, $a3, $67, $2d,
   $4b, $be, $e6, $19, $51, $5f, $9f, $05, $08, $78, $c4, $4a, $66, $f5, $58, $ff,
   $48, $0e, $c0, $9a, $0d, $70, $bc, $8e, $2c, $93, $ad, $a7, $b7, $46, $ce, $5a,
   $97, $7d, $cc, $32, $a2, $bf, $3e, $0a, $10, $f1, $88, $94, $cd, $ea, $b1, $fe,
   $90, $1d, $81, $34, $1a, $e1, $79, $1c, $59, $27, $5b, $4f, $6e, $8d, $9c, $b5,
   $2e, $fb, $98, $65, $45, $7e, $7c, $14, $21, $e3, $11, $29, $9b, $d5, $63, $fd,
) ;



(* Tables for RS decoder */
/* Galois field log/antilog tables *)
ALPHA_TO : array[0..255] of byte =
  (
    $01, $02, $04, $08, $10, $20, $40, $80, $87, $89, $95, $ad, $dd, $3d, $7a, $f4,
    $6f, $de, $3b, $76, $ec, $5f, $be, $fb, $71, $e2, $43, $86, $8b, $91, $a5, $cd,
    $1d, $3a, $74, $e8, $57, $ae, $db, $31, $62, $c4, $0f, $1e, $3c, $78, $f0, $67,
    $ce, $1b, $36, $6c, $d8, $37, $6e, $dc, $3f, $7e, $fc, $7f, $fe, $7b, $f6, $6b,
    $d6, $2b, $56, $ac, $df, $39, $72, $e4, $4f, $9e, $bb, $f1, $65, $ca, $13, $26,
    $4c, $98, $b7, $e9, $55, $aa, $d3, $21, $42, $84, $8f, $99, $b5, $ed, $5d, $ba,
    $f3, $61, $c2, $03, $06, $0c, $18, $30, $60, $c0, $07, $0e, $1c, $38, $70, $e0,
    $47, $8e, $9b, $b1, $e5, $4d, $9a, $b3, $e1, $45, $8a, $93, $a1, $c5, $0d, $1a,
    $34, $68, $d0, $27, $4e, $9c, $bf, $f9, $75, $ea, $53, $a6, $cb, $11, $22, $44,
    $88, $97, $a9, $d5, $2d, $5a, $b4, $ef, $59, $b2, $e3, $41, $82, $83, $81, $85,
    $8d, $9d, $bd, $fd, $7d, $fa, $73, $e6, $4b, $96, $ab, $d1, $25, $4a, $94, $af,
    $d9, $35, $6a, $d4, $2f, $5e, $bc, $ff, $79, $f2, $63, $c6, $0b, $16, $2c, $58,
    $b0, $e7, $49, $92, $a3, $c1, $05, $0a, $14, $28, $50, $a0, $c7, $09, $12, $24,
    $48, $90, $a7, $c9, $15, $2a, $54, $a8, $d7, $29, $52, $a4, $cf, $19, $32, $64,
    $c8, $17, $2e, $5c, $b8, $f7, $69, $d2, $23, $46, $8c, $9f, $b9, $f5, $6d, $da,
    $33, $66, $cc, $1f, $3e, $7c, $f8, $77, $ee, $5b, $b6, $eb, $51, $a2, $c3, $00,
 ) ;

INDEX_OF : array[0..255] of byte =
  (
   $ff, $00, $01, $63, $02, $c6, $64, $6a, $03, $cd, $c7, $bc, $65, $7e, $6b, $2a,
   $04, $8d, $ce, $4e, $c8, $d4, $bd, $e1, $66, $dd, $7f, $31, $6c, $20, $2b, $f3,
   $05, $57, $8e, $e8, $cf, $ac, $4f, $83, $c9, $d9, $d5, $41, $be, $94, $e2, $b4,
   $67, $27, $de, $f0, $80, $b1, $32, $35, $6d, $45, $21, $12, $2c, $0d, $f4, $38,
   $06, $9b, $58, $1a, $8f, $79, $e9, $70, $d0, $c2, $ad, $a8, $50, $75, $84, $48,
   $ca, $fc, $da, $8a, $d6, $54, $42, $24, $bf, $98, $95, $f9, $e3, $5e, $b5, $15,
   $68, $61, $28, $ba, $df, $4c, $f1, $2f, $81, $e6, $b2, $3f, $33, $ee, $36, $10,
   $6e, $18, $46, $a6, $22, $88, $13, $f7, $2d, $b8, $0e, $3d, $f5, $a4, $39, $3b,
   $07, $9e, $9c, $9d, $59, $9f, $1b, $08, $90, $09, $7a, $1c, $ea, $a0, $71, $5a,
   $d1, $1d, $c3, $7b, $ae, $0a, $a9, $91, $51, $5b, $76, $72, $85, $a1, $49, $eb,
   $cb, $7c, $fd, $c4, $db, $1e, $8b, $d2, $d7, $92, $55, $aa, $43, $0b, $25, $af,
   $c0, $73, $99, $77, $96, $5c, $fa, $52, $e4, $ec, $5f, $4a, $b6, $a2, $16, $86,
   $69, $c5, $62, $fe, $29, $7d, $bb, $cc, $e0, $d3, $4d, $8c, $f2, $1f, $30, $dc,
   $82, $ab, $e7, $56, $b3, $93, $40, $d8, $34, $b0, $ef, $26, $37, $0c, $11, $44,
   $6f, $78, $19, $9a, $47, $74, $a7, $c1, $23, $53, $89, $fb, $14, $5d, $f8, $97,
   $2e, $4b, $b9, $60, $0f, $ed, $3e, $e5, $f6, $87, $a5, $17, $3a, $a3, $3c, $b7,

  ) ;

// -----------------------------------------------------------------------------------

(*
/* --------------- */
/* Viterbi decoder */
/* --------------- *)

(* Viterbi decoder for arbitrary convolutional code
 * viterbi27 and viterbi37 for the r=1/2 and r=1/3 K=7 codes are faster
 * Copyright 1997 Phil Karn, KA9Q
 *)

(* This is a bare bones <2,7> Viterbi decoder, adapted from a general purpose model.
 * It is not optimised in any way, neither as to coding (for example the memcopy should
 * be achievable simply by swapping pointers), nor as to simplifying the metric table,
 * nor as to using any machine-specific smarts.  On contemporary machines, in this application,
 * the execution time is negligible.  Many ideas for optimisation are contained in PK's www pages.
 * The real ADC is 8-bit, though in practice 4-bits are actually sufficient.
 * Descriptions of the Viterbi decoder algorithm can be found in virtually any book
 * entitled "Digital Communications".  (JRM)
 */
*)

function viterbi27(data : ^byte;          // Decoded output data
                   symbols : ^byte;       // Raw deinterleaved input symbols
                   nbits : Cardinal)      // Number of output bits
                  : integer ;

var

     bitcnt : Cardinal ;
     beststate,i,j : integer ;
     cmetric : array[0..63] of integer ;
     nmetric : array[0..63] of integer ;    // 2^(K-1)
     pp : ^Cardinal ;
     m0,m1,mask : integer ;
     mets : array[0..3] of integer ;        // 2^N
     unsigned long *paths;
     paths : array[0..nbits*2-1] of Cardinal ;
     b1,b2 : integer ;

begin

     bitcnt := 0 ;

//   pp = paths = (unsigned long *)calloc(nbits*2,sizeof(unsigned long));

     pp := 0 ;

//   Initialize starting metrics to prefer 0 state

     cmetric[0] := 0 ;

     for i := 1 to 63 do begin // check !!!!
         cmetric[i] := -999999 ;
     end ;

     repeat
           // Read 2 input symbols and compute the 4 branch metrics
           for i := 0 to 3 do begin
               mets[i] := 0 ;
               for j := 0 to 1 do begin
                   Inc(mets[i], mettab[(i shr (1-j)) and 1][symbols[j]]) ;
               end ;
           end ;
           Inc(symbols,2) ;
           mask := 1 ;
           i := 0 ;
           while (i<64) do begin
               b1 := mets[Syms[i]] ;
               m0 := cmetric[i/2] + b1 ;
               nmetric[i] := m0 ;
               b2 := mets[Syms[i+1]] ;
               Dec(b1,b2) ;
               m1 := cmetric[(i div 2) + (1 shl (K-2))] + b2;
               if (m1 > m0) then begin
                  nmetric[i] := m1 ;
                  paths[pp] := paths[pp] or mask ;
               end ;
               Dec(m0,b1) ;
               nmetric[i+1] := m0 ;
               Inc(m1,b1) ;
               if (m1 > m0) then begin
                   nmetric[i+1] := m1 ;
                   paths[pp] := paths[pp] or mask shl 1 ;
               end ;
               mask := mask shl 2 ;
               if (mask = 0) then begin
                   mask = 1;
                   Inc(pp) ;
               end ;
               Inc(i,2) ;
           end ;
           if (mask <> 1) then begin
               Inc(pp);
           end ;
           Inc(bitcnt) ;
           if (bitcnt = nbits) then begin
               beststate := 0 ;
               break ;
           end ;
           memcpy(cmetric,nmetric,sizeof(cmetric));
     until false ;
     Dec(pp,2) ;
     // Chain back from terminal state to produce decoded data
     memset(data,0,nbits/8);
     for i := nbits-K downto 0 do begin
         if ((pp[beststate shr 5] and (1L shl (beststate and 31))) <>0) then begin
             beststate := beststate or (1 shl (K-1)) ;
             data[i shr 3] := data[i shr 3] or $80 shr (i and 7);
         end ;
         beststate := beststate shr 1 ;
         Dec(pp,2) ;
     end ;
     result := 0 ;
end { viterbi27 } ;

// -----------------------------------------------------------------------------------

(*
/* ---------- */
/* RS Decoder */
/* ---------- */

/* This decoder has evolved extensively through the work of Phil Karn.  It draws
 * on his own ideas and optimisations, and on the work of others.  The lineage
 * is as below, and parts of the authors' notices are included here.  (JRM)

 * Reed-Solomon decoder
 * Copyright 2002 Phil Karn, KA9Q
 * May be used under the terms of the GNU General Public License (GPL)
 *
 * Reed-Solomon coding and decoding
 * Phil Karn (karn@ka9q.ampr.org) September 1996
 *
 * This file is derived from the program "new_rs_erasures.c" by Robert
 * Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari Thirumoorthy
 * (harit@spectra.eng.hawaii.edu), Aug 1995
 * --------------------------------------------------------------------------
 *
 * From the RM-Z & HT program:
 * The encoding and decoding methods are based on the
 * book "Error Control Coding: Fundamentals and Applications",
 * by Lin and Costello, Prentice Hall, 1983, ISBN 0-13-283796-X
 * Portions of this program are from a Reed-Solomon encoder/decoder
 * in C, written by Simon Rockliff (simon@augean.ua.oz.au) on 21/9/89.
 * --------------------------------------------------------------------------
 *
 * From the 1989/1991 SR program (also based on Lin and Costello):
 * This program may be freely modified and/or given to whoever wants it.
 * A condition of such distribution is that the author's contribution be
 * acknowledged by his name being left in the comments heading the program,
 *                               Simon Rockliff, 26th June 1991
 *
 *)

function mod255(x:integer):integer ;
begin
     while (x >= 255) do begin
         Dec(x,255) ;
         x := (x shr 8) + (x and 255);
     end ;
     result := x ;
end { mod255 } ;

function decode_rs_8(char *data, int *eras_pos, int no_eras) : integer ;
var
     deg_lambda, el, deg_omega : integer ;
     i, j, r, k : integer ;
     u,q,tmp,num1,num2,den,discr_r : byte ;
     lambda : array [0..NROOTS] of byte ;
     s : array [0..NROOTS-1] of byte ;   // Err+Eras Locator poly and syndrome poly

     unsigned char b[NROOTS+1], t[NROOTS+1], omega[NROOTS+1];
     unsigned char root[NROOTS], reg[NROOTS+1], loc[NROOTS];
     syn_error, count : integer ;
begin

// form the syndromes; i.e., evaluate data(x) at roots of g(x)

     for i := 0 to NROOTS-1 do begin
         s[i] := data[0] ;
     end ;

     for j :=1 to NN-1 do begin
         for i := 0 to NROOTS-1 do begin
             if (s[i] = 0) then begin
                 s[i] := data[j] ;
             end
             else begin
                 s[i] := data[j] xor ALPHA_TO[mod255(INDEX_OF[s[i]] + (FCR+i)*PRIM)] ;
             end ;
          end ;
     end ;

// Convert syndromes to index form, checking for nonzero condition */

     syn_error := 0 ;
     for i := 0 to NROOTS-1 do begin
         syn_error := syn_error or s[i] ;
         s[i] := INDEX_OF[s[i]] ;
     end ;

     if (syn_error=0) then begin
    (* if syndrome is zero, data[] is a codeword and there are no
     * errors to correct. So return data[] unmodified
     *)
        count := 0 ;
        goto finish ;
     end ;

  memset(&lambda[1],0,NROOTS*sizeof(lambda[0]));
  lambda[0] = 1;

  if (no_eras > 0) then begin
    // Init lambda to be the erasure locator polynomial
      lambda[1] := ALPHA_TO[mod255(PRIM*(NN-1-eras_pos[0]))] ;
      for i := 1 to no_eras-1 do begin
          u = mod255(PRIM*(NN-1-eras_pos[i]));
          for j := i+1 downto 0 do begin
              tmp := INDEX_OF[lambda[j - 1]] ;
              if (tmp <> A0) then begin
                  lambda[j] := lambda[j] xor ALPHA_TO[mod255(u + tmp)] ;
              end ;
          end ;
      end ;
  end ;

  for(i=0;i<NROOTS+1;i++)
    b[i] = INDEX_OF[lambda[i]];

  (*
   * Begin Berlekamp-Massey algorithm to determine error+erasure
   * locator polynomial
   *)
  r = no_eras;
  el = no_eras;
  while (++r <= NROOTS) do begin       // r is the step number
    // Compute discrepancy at the r-th step in poly-form
    discr_r = 0;
    for (i = 0; i < r; i++) do begin
      if ((lambda[i] != 0) && (s[r-i-1] != A0)) then begin
        discr_r ^= ALPHA_TO[mod255(INDEX_OF[lambda[i]] + s[r-i-1])];
      end ;
    end ;
    discr_r = INDEX_OF[discr_r];        // Index form
    if (discr_r == A0) then begin
      // 2 lines below: B(x) <-- x*B(x)
      memmove(&b[1],b,NROOTS*sizeof(b[0]));
      b[0] = A0;
    end
    else begin
      // 7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x)
      t[0] = lambda[0];
      for (i = 0 ; i < NROOTS; i++) do begin
        if(b[i] != A0)
          t[i+1] = lambda[i+1] ^ ALPHA_TO[mod255(discr_r + b[i])];
        else
          t[i+1] = lambda[i+1];
      end ;
      if (2 * el <= r + no_eras - 1) then begin
        el = r + no_eras - el;
        (*
         * 2 lines below: B(x) <-- inv(discr_r) *
         * lambda(x)
         *)
        for (i = 0; i <= NROOTS; i++) do begin
          b[i] = (lambda[i] == 0) ? A0 : mod255(INDEX_OF[lambda[i]] - discr_r + NN);
        end ;
      end
      else begin
        /* 2 lines below: B(x) <-- x*B(x) */
        memmove(&b[1],b,NROOTS*sizeof(b[0]));
        b[0] = A0;
      end ;
      memcpy(lambda,t,(NROOTS+1)*sizeof(t[0]));
    end ;
  }

  /* Convert lambda to index form and compute deg(lambda(x)) */
  deg_lambda = 0;
  for(i=0;i<NROOTS+1;i++)do begin
    lambda[i] = INDEX_OF[lambda[i]];
    if(lambda[i] <> A0) then begin
      deg_lambda = i;
    end ;

  end ;
  // Find roots of the error+erasure locator polynomial by Chien search
  memcpy(&reg[1],&lambda[1],NROOTS*sizeof(reg[0]));
  count = 0;            /* Number of roots of lambda(x) */
  for (i = 1,k=IPRIM-1; i <= NN; i++,k = mod255(k+IPRIM)) do begin
    q = 1; /* lambda[0] is always 0 */
    for (j = deg_lambda; j > 0; j--)do begin
      if (reg[j] != A0) then begin
        reg[j] = mod255(reg[j] + j);
        q ^= ALPHA_TO[reg[j]];
      end ;
    end ;
    if (q != 0)
      continue; /* Not a root */
    /* store root (index-form) and error location number */
    root[count] = i;
    loc[count] = k;
    /* If we've already found max possible roots,
     * abort the search to save time
     */
    if(++count == deg_lambda) then begin
      break;
    end ;
  end ;
  if (deg_lambda != count) then begin
    (*
     * deg(lambda) unequal to number of roots => uncorrectable
     * error detected
     *)
    count = -1;
    goto finish;
  end ;
  (*
   * Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
   * x**NROOTS). in index form. Also find deg(omega).
   *)
  deg_omega = 0;
  for (i = 0; i < NROOTS;i++)do begin
    tmp = 0;
    j = (deg_lambda < i) ? deg_lambda : i;
    for(;j >= 0; j--) do begin
      if ((s[i - j] != A0) && (lambda[j] != A0)) then begin
        tmp ^= ALPHA_TO[mod255(s[i - j] + lambda[j])];
      end ;
    end ;
    if(tmp != 0) then begin
      deg_omega = i;
    end ;
    omega[i] = INDEX_OF[tmp];
  end ;
  omega[NROOTS] = A0;

  (*
   * Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
   * inv(X(l))**(FCR-1) and den = lambda_pr(inv(X(l))) all in poly-form
   *)
  for (j = count-1; j >=0; j--) do begin
    num1 = 0;
    for (i = deg_omega; i >= 0; i--) do begin
      if (omega[i] != A0) then begin
        num1  ^= ALPHA_TO[mod255(omega[i] + i * root[j])];
      end ;
    end ;
    num2 = ALPHA_TO[mod255(root[j] * (FCR - 1) + NN)];
    den = 0;

    // lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i]
    for (i = minm(deg_lambda,NROOTS-1) & ~1; i >= 0; i -=2) do begin
      if(lambda[i+1] != A0) then begin
        den ^= ALPHA_TO[mod255(lambda[i+1] + i * root[j])];
      end ;
    end ;
    if (den = 0) then begin
      count = -1;
      goto finish;
    end ;
    // Apply error to data
    if (num1 <> 0) then begin
      data[loc[j]] ^= ALPHA_TO[mod255(INDEX_OF[num1] + INDEX_OF[num2] + NN - INDEX_OF[den])];
    end ;
  end ;
finish:
  if(eras_pos != NULL) then begin
    for(i=0;i<count;i++) do begin
      eras_pos[i] = loc[i];
    end ;
  end ;
  result := count ;
end { decode_rs_8 } ;

// -----------------------------------------------------------------------------------

(* ---------- */
/* Re-encoder */
/* ---------- */

/* Reference encoder for proposed coded AO-40 telemetry format - v1.0  7 Jan 2002
 * Copyright 2002, Phil Karn, KA9Q
 * This software may be used under the terms of the GNU Public License (GPL)
 */

/* Adapted from  the above enc_ref.c  as used by the spacecraft (JRM) *)


// Write one binary channel symbol into the interleaver frame and update the pointers

procedure interleave_symbol(c:integer)
var
     row, col : integer ;
begin
     col := Bindex div COLUMNS ;
     row := Bindex mod COLUMNS ;
     if (c<>0) then begin
         reencode[row*ROWS+col] := 1 ;
     end ;
     Inc(Bindex) ;
end { interleave_symbol } ;

// Convolutionally encode and interleave one byte
procedure encode_and_interleave(c:byte;cnt:integer) ;
begin
     while (cnt <> 0) do begin
         Dec(cnt) ;
         Conv_sr := (Conv_sr shl 1) or (c shr 7) ;
         c := c shl 1 ;
         interleave_symbol( Partab[Conv_sr and CPOLYA]) ;
         interleave_symbol(not Partab[Conv_sr and CPOLYB]) ; // Second encoder symbol is inverted
     end ;
end { encode_and_interleave } ;

// Scramble a byte, convolutionally encode and interleave into frame
procedure scramble_and_encode(c:byte) ;
begin
     c := c xor Scrambler[Nbytes] ;  // Scramble byte
     encode_and_interleave(c,8);     // RS encode and place into reencode buffer
end { scramble_and_encode } ;

(* Three user's entry points now follow.  They are:

   init_encoder()                   Called once before using system.
   encode_byte(unsigned char c)     Called with each user byte (i.e. 256 calls)
   encode_parity()                  Called 64 times to finish off
*)

// This function initializes the encoder.
procedure init_encoder ;

var

  i, j, sr : integer ;

begin

     Nbytes  := 0 ;
     Conv_sr := 0 ;
     Bindex  := COLUMNS ;              // Sync vector is in first column; data starts here

     for j := 0 to RSBLOCKS-1 do begin   // Flush parity array
         for i :=0 to NROOTS-1 do begin
             RS_block[j][i] := 0 ;
         end ;
     end ;

//   Clear re-encoded array

     for i :=0 to 5200-1 do begin
         reencode[i] := 0 ;
     end ;

//   Generate sync vector, interleave into re-encode array, 1st column

     sr := $7f ;
     for i := 0 to 65 do begin
        if (sr and 64 <> 0)
           reencode[ROWS*i] := 1 ;      // Every 80th symbol is a sync bit
           sr := (sr shl 1) or Partab[sr and SYNC_POLY] ;
     end ;

end { init_encoder } ;


(* This function is called with each user data byte to be encoded into the
 * current frame. It should be called in sequence 256 times per frame, followed
 * by 64 calls to encode_parity().
 *)

procedure encode_byte(c:byte) ;
var
     unsigned char *rp;                // RS block pointer
     int i;
     unsigned char feedback;
begin

// Update the appropriate Reed-Solomon codeword

     rp := RS_block[Nbytes and 1] ;

// Compute feedback term

     feedback = INDEX_OF[c xor rp[0]] ;

(* If feedback is non-zero, multiply by each generator polynomial coefficient and
   * add to corresponding shift register elements
   *)

  if(feedback != A0) then begin
    int j;

    (* This loop exploits the palindromic nature of the generator polynomial
     * to halve the number of discrete multiplications
     *)
    for(j=0;j<15;j++) do begin
      unsigned char t;

      t = ALPHA_TO[mod255(feedback + RS_poly[j])];
      rp[j+1] ^= t; rp[31-j] ^= t;
    end ;
    rp[16] ^= ALPHA_TO[mod255(feedback + RS_poly[15])];
  end ;

  // Shift 32 byte RS register one position down
  for(i=0;i<31;i++) do begin
    rp[i] = rp[i+1];
  end ;
  // Handle highest order coefficient, which is unity
  if(feedback != A0) then begin
    rp[31] = ALPHA_TO[feedback];
  end
  else begin
    rp[31] = 0;
  end ;
  scramble_and_encode(c);
  Inc(Nbytes) ;
end { encode_byte } ;

(* This function should be called 64 times after the 256 data bytes
 * have been passed to update_encoder. Each call scrambles, encodes and
 * interleaves one byte of Reed-Solomon parity.
 *)

procedure encode_parity ;
var
     c : byte ;
begin
     c :=  RS_block[Nbytes and 1][(Nbytes-256) shr 1] ;
     scramble_and_encode(c) ;
     Inc(NBytes) ;
     if (Nbytes = 320) then begin
         // Tail off the convolutional encoder (flush)
         encode_and_interleave(0,6) ;
     end ;
end { encode_parity } ;


(* Encodes the 256 byte source block RSdecdata[] into 5200 byte block of symbols
 * reencode[].  It has the same format as an off-air received block of symbols.
 *)

procedure encode_FEC40(var reencode  : TEncodedSymbols;        // Re-encoded symbols
                       var RSdecdata : TRSDecoderDataOutput) ; // User's source data
var
     i : integer ;
begin

     init_encoder ;

     for i := 0 to 255 do begin
         encode_byte(RSdecdata[i]) ;
     end ;

     for i := 0 to 63 do begin
         encode_parity() ;
     end ;

end { encode_FEC40 } ;


// -----------------------------------------------------------------------------------

(****************************************************************************************************/
/*                                                                                                  */
/*   AMSAT AO-40 FEC Reference Decoder (by JRM)               Input                Output           */
/*   ------------------------------------------               -----------------------------------   */
/*   Step 1: De-interleave                                    raw[5200]       ->  symbols[5132]     */
/*   Step 2: Viterbi decoder                                  symbols[5132]   ->  vitdecdata[320]   */
/*   Step 3: RS decoder                                       vitdecdata[320] ->  RSdecdata[256]    */
/*   Step 4: Option: Re-encode o/p and count channel errors   RSdecdata[256]  ->  reencode[5200]    */
/*                                                                                                  */
/*                This main() has been modified by VK5HI for evaluation purposes.                   */
/*                      James's original main() follows as main_Original()                          */
/*                                                                                                  */
/****************************************************************************************************)

#include <math.h>	                             /* for sqrt */

int main(int argc, char *argv[])
{
        unsigned char raw[SYMPBLOCK] ;           /* 5200 raw received;  sync+symbols  */
        unsigned char symbols[NBITS*2+65+3] ;    /* de-interleaved sync+symbols */
        unsigned char vitdecdata[(NBITS-6)/8] ;  /* array for Viterbi decoder output data */
        unsigned char RSdecdata[BLOCKSIZE] ;     /* RS decoder data output (to user)*/

unsigned char sync[] = {1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,1,
						1,1,1,0,0,1,0,1,1,0,0,1,0,0,1,0,
						0,0,0,0,0,1,0,0,0,1,0,0,1,1,0,0,
						0,1,0,1,1,1,0,1,0,1,1,0,1,1,0,0,0};

		 FILE *InFile, *OutFile;
unsigned char *inpBuf, *inBuf, bHour, bMin, bSec, chX, chY, chZ, sync_bit;
		 char achFileName[80];
		 int x, y, z, count = 0;
		 long filesize, memact, memsize, loop, loopcnt, loopstart;
		 long totalerrors = 0, ptr = 0;
		 FILE *fp ;								/* Needed if you save o/p to file */
		 int i, sym, sync_volts = 0, sync_gain;
		 long frame_energy = 1;

		 if (argc != 2) {
			 printf("Enter FECDecode [Input .FEC File]\n");
			 return(0);
		 }

		/* open input file */
		if ((InFile=fopen(argv[1], "rb")) == NULL) {
			fprintf(stderr, "cannot open Input File. \n");
			return (1);
		}
		filesize = filelength(fileno(InFile));
		memsize = (1 + filesize / 8192) * 8192;
		//Establish the number of block loops
		loop = filesize / 650;
		//allocate memory
		inpBuf = (unsigned char *)malloc(memsize*sizeof(unsigned char));
		if (inpBuf == NULL) return(0);
		/* clear buffer */
		memset(inpBuf, 0x00, memsize*sizeof(unsigned char));
		/* read input file into buffer */
		x=0;
		while (!feof(InFile)) {
			inpBuf[x] = fgetc(InFile);
			x++;
		}
		fclose(InFile);

		// Now allocate memory to hold the reconstructed symbols
		// ascertain size of data which is also ptr to end of data
		memact = filesize / 650 * SYMPBLOCK;
		// allocate memory size
		memsize = (1 + memact / 8192) * 8192;
		inBuf = (unsigned char *)malloc(memsize*sizeof(unsigned char));
		if (inBuf == NULL) {
			free (inpBuf);
			return(1);
		}
		/* clear buffer */
		memset(inBuf, 0x00, memsize*sizeof(unsigned char));
		z=0;
		//create "fudged" symbols
		for (loopcnt=0; loopcnt<loop; loopcnt++){
			for (count=0; count<650; count++){
				chX=inpBuf[count+loopcnt*650];
				for (y=0; y<8; y++){
					if (y==0){
						chZ = chX & 0x80;
					} else {
						chY=_rotl(chX,y);
						chZ = chY & 0x80;
					}
					if (chZ == 0x80) {
						inBuf[z] = 0xf0;
					} else {
						inBuf[z] = 0x10;
					}
					z++;
				}// for y
			}//for count
		}//for loopcnt
		/* free up Input Buffer */
		free(inpBuf);
		fprintf(stderr, "Searching for valid FEC Blocks. \n");
		/* Lets now do a test to establish whether blocks are truly FEC */
		/* Loop till we have sync detection */
		/* Now search for sync_vector using code by G3RUH (24 June 03) */
		do {
			for (i=0; i<65; i++) {
				sync_bit = sync[i];
				//if ptr is pointing to valid memory then get symbol.
				if (ptr < memact) sym = inBuf[ptr] - 128;
				if (sync_bit) sync_volts += sym; else sync_volts -= sym;
				frame_energy += sym * sym;
				ptr = ptr + 80;
			}
			count++;
			ptr = count;
			sync_gain = sync_volts/(int)sqrt(frame_energy/65);
			if (sync_gain > 41) goto EXIT_LOOP;
			sync_volts=0; frame_energy = 1;
		} while (ptr < memact);
EXIT_LOOP:
		// if no valid blocks then exit program
		if (ptr >= memact) {
			free(inBuf);
			fprintf(stderr, "No valid FEC Blocks detected in this file. \n");
			//rename file to .nbg
			memset(achFileName, 0x00, sizeof(achFileName));
			strncpy(achFileName, argv[1], strlen(argv[1])-3);
			strcat(achFileName, "nbg");
			if (rename(argv[1], achFileName) !=0) {
				fprintf(stderr, "Cannot rename .FEC file. \n");
			}
			return (1);
		}
		// otherwise continue processing from first valid block
		loopstart = ptr / SYMPBLOCK;
		/* open output txt file  for verbose information*/
		memset(achFileName, 0x00, sizeof(achFileName));
		strncpy(achFileName, argv[1], strlen(argv[1])-3);
		strcat(achFileName, "txt");
		if ((OutFile=fopen(achFileName, "wb")) == NULL) {
			fprintf(stderr, "cannot open Output File. \n");
			return (1);
		}
		/* rename previous .fec to .old */
		memset(achFileName, 0x00, sizeof(achFileName));
		strncpy(achFileName, argv[1], strlen(argv[1])-3);
		strcat(achFileName, "old");
		if (rename(argv[1], achFileName) !=0) {
			fprintf(stderr, "Cannot rename .FEC file. \n");
			return (1);
		};
		/* open output fec file */
		memset(achFileName, 0x00, sizeof(achFileName));
		strncpy(achFileName, argv[1], strlen(argv[1])-3);
		strcat(achFileName, "fec");
	    fp=fopen(achFileName,"wb");                       /* Output 256 bytes of user's data */
		/* now lets do some serious crunching */
		for (count=loopstart; count<loop; count++){
			for (x=0; x<5200; x++){
				raw[x] = inBuf[x + count*5200];
			}
		/* Step 1: De-interleaver */
			  {
				/* Input  array:  raw   */
				/* Output array:  symbols */
				int col,row;
				int coltop,rowstart;

				coltop=0;
				for(col=1;col<ROWS;col++){          /* Skip first column as it's the sync vector */
				  rowstart=0;
				  for(row=0;row<COLUMNS;row++){
					symbols[coltop+row]=raw[rowstart+col];  /* coltop=col*65 ; rowstart=row*80 */
					rowstart+=ROWS;
				  }
				  coltop+=COLUMNS;
				}
			  } /* end of de-interleaving */


		/* Step 2: Viterbi decoder */
		/* ----------------------- */
			  {
				/* Input  array:  symbols  */
				/* Output array:  vitdecdata */
				viterbi27(vitdecdata,symbols,NBITS);
			  }

		/* Steps 3: RS decoder */
		/* ------------------- */

		/* There are two RS decoders, processing 128 bytes each.
		 *
		 * If both RS decoders are SUCCESSFUL
		 * On exit:
		 *   rs_failures = 0
		 *   rserrs[x]   = number of errors corrected;  range 0 to 16   (x= 0 or 1)
		 *   Data output is in array RSdecdata[256].
		 *
		 * If an RS decoder FAILS
		 * On exit:
		 *   rs_failures = 1 or 2   (i.e. != 0)
		 *   rserrs[x] contains -1
		 *   Data output should not be used.
		 */

			  {
				/* Input  array:  vitdecdata */
				/* Output array:  RSdecdata  */

				char rsblocks[RSBLOCKS][NN];          /*  [2][255] */
				int row,col,di,si;
				int rs_failures ;                     /* Flag set if errors found */
				int rserrs[RSBLOCKS];                 /* Count of errors in each RS codeword */


				/* Interleave into Reed Solomon codeblocks */
				memset(rsblocks,0,sizeof(rsblocks));                          /* Zero rsblocks array */
				di = 0;
				si = 0;
				for(col=RSPAD;col<NN;col++){
				  for(row=0;row<RSBLOCKS;row++){
					rsblocks[row][col] = vitdecdata[di++] ^ Scrambler[si++];  /* Remove scrambling */
				  }
				}
				/* Run RS-decoder(s) */
				rs_failures = 0;
				for(row=0;row<RSBLOCKS;row++){
				  rserrs[row] = decode_rs_8(rsblocks[row],NULL,0);
				  rs_failures += (rserrs[row] == -1);
				}

				/* If frame decoded OK, deinterleave data from RS codeword(s), and save file */
				bHour=0; bMin=0; bSec=0;
				if(!rs_failures){
				  int j = 0;

				  for(col=RSPAD;col<KK;col++){
					for(row=0;row<RSBLOCKS;row++){
					  RSdecdata[j++] = rsblocks[row][col];
					}
				  }
				  /* and save out succesfully RS-decoded data */
				  fwrite(RSdecdata,1,sizeof(RSdecdata),fp);
					memcpy(&bHour, RSdecdata+0x00ab, 1);
					memcpy(&bMin, RSdecdata+0x00aa, 1);
					memcpy(&bSec, RSdecdata+0x00a9, 1);
				}

				/* Print RS-decode status summary */
				if(Verbose){
				  fprintf(stderr,"Block %3d Time = %02d:%02d:%02d RS byte corrections: ",count+1, bHour,bMin, bSec);
				  fprintf(OutFile,"Block %3d Time = %02d:%02d:%02d RS byte corrections: ", count+1, bHour, bMin, bSec);
				  for(row=0;row<RSBLOCKS;row++){
					  if(rserrs[row] != -1){
						  fprintf(stderr," %2d  ",rserrs[row]);
						  fprintf(OutFile," %2d  ",rserrs[row]);
					  }else{
						  fprintf(stderr,"FAIL ");
						  fprintf(OutFile,"FAIL ");
					  }
				  }
				}
			  } /* end of rs section */


		/* Step 4: Optional: Re-encode o/p and count channel errors  */
		/* --------------------------------------------------------  */

			 {  int errors = 0;
				int i;
				/* Input  array:  RSdecdata */
				/* Output array:  reencode  */

				encode_FEC40(reencode,RSdecdata) ;  /* Re-encode in AO-40 FEC format */

				/* Count the channel errors */
				errors = 0;
				for(i=0;i<SYMPBLOCK;i++)
				  if ( reencode[i] != (raw[i]>>7) ) {
					errors++ ;
				  }

				if(Verbose){
				  fprintf(stderr," Channel symbol errors: %3d (%.3g%%)\n",errors,100.*errors/SYMPBLOCK);
				  fprintf(OutFile," Channel symbol errors: %3d (%.3g%%)\r\n",errors,100.*errors/SYMPBLOCK);
				}
				totalerrors = totalerrors + errors;
			 }  /* end of reencode section */

		/* --------------------------------------------------------*/

	 } // for count
     fprintf(stderr,"Total Symbol Errors %2d  ",totalerrors);
     fprintf(OutFile,"Total Symbol Errors %2d  ",totalerrors);
     fclose(fp);
 	 fclose(OutFile);
	 free(inBuf);
	 return 0;
}   /*end of main*/

/****************************************************************************************************/
/*                                                                                                  */
/*   AMSAT AO-40 FEC Reference Decoder (by JRM)               Input                Output           */
/*   ------------------------------------------               -----------------------------------   */
/*   Step 1: De-interleave                                    raw[5200]       ->  symbols[5132]     */
/*   Step 2: Viterbi decoder                                  symbols[5132]   ->  vitdecdata[320]   */
/*   Step 3: RS decoder                                       vitdecdata[320] ->  RSdecdata[256]    */
/*   Step 4: Option: Re-encode o/p and count channel errors   RSdecdata[256]  ->  reencode[5200]    */
/*                                                                                                  */
/****************************************************************************************************/


int main_Original()
{
        unsigned char raw[SYMPBLOCK] ;           /* 5200 raw received;  sync+symbols  */
        unsigned char symbols[NBITS*2+65+3] ;    /* de-interleaved sync+symbols */
        unsigned char vitdecdata[(NBITS-6)/8] ;  /* array for Viterbi decoder output data */
        unsigned char RSdecdata[BLOCKSIZE] ;     /* RS decoder data output (to user)*/

/* Step 1: De-interleaver */
      {
        /* Input  array:  raw   */
        /* Output array:  symbols */
        int col,row;
        int coltop,rowstart;
        FILE *fp ;                          /* File pointer */

        fp=fopen("fec","r");                /* Input 5200 symbols (incl sync) to data array */
        fread(raw,1, sizeof(raw),fp);
        fclose(fp);

        coltop=0;
        for(col=1;col<ROWS;col++){          /* Skip first column as it's the sync vector */
          rowstart=0;
          for(row=0;row<COLUMNS;row++){
            symbols[coltop+row]=raw[rowstart+col];  /* coltop=col*65 ; rowstart=row*80 */
            rowstart+=ROWS;
          }
          coltop+=COLUMNS;
        }
      } /* end of de-interleaving */


/* Step 2: Viterbi decoder */
/* ----------------------- */
      {
        /* Input  array:  symbols  */
        /* Output array:  vitdecdata */
        viterbi27(vitdecdata,symbols,NBITS);
      }

/* Steps 3: RS decoder */
/* ------------------- */

/* There are two RS decoders, processing 128 bytes each.
 *
 * If both RS decoders are SUCCESSFUL
 * On exit:
 *   rs_failures = 0
 *   rserrs[x]   = number of errors corrected;  range 0 to 16   (x= 0 or 1)
 *   Data output is in array RSdecdata[256].
 *
 * If an RS decoder FAILS
 * On exit:
 *   rs_failures = 1 or 2   (i.e. != 0)
 *   rserrs[x] contains -1
 *   Data output should not be used.
 */

      {
        /* Input  array:  vitdecdata */
        /* Output array:  RSdecdata  */

        char rsblocks[RSBLOCKS][NN];          /*  [2][255] */
        int row,col,di,si;
        int rs_failures ;                     /* Flag set if errors found */
        int rserrs[RSBLOCKS];                 /* Count of errors in each RS codeword */

        FILE *fp ;                            /* Needed if you save o/p to file */

        /* Interleave into Reed Solomon codeblocks */
        memset(rsblocks,0,sizeof(rsblocks));                          /* Zero rsblocks array */
        di = 0;
        si = 0;
        for(col=RSPAD;col<NN;col++){
          for(row=0;row<RSBLOCKS;row++){
            rsblocks[row][col] = vitdecdata[di++] ^ Scrambler[si++];  /* Remove scrambling */
          }
        }
        /* Run RS-decoder(s) */
        rs_failures = 0;
        for(row=0;row<RSBLOCKS;row++){
          rserrs[row] = decode_rs_8(rsblocks[row],NULL,0);
          rs_failures += (rserrs[row] == -1);
        }

        /* If frame decoded OK, deinterleave data from RS codeword(s), and save file */
        if(!rs_failures){
          int j = 0;

          for(col=RSPAD;col<KK;col++){
            for(row=0;row<RSBLOCKS;row++){
              RSdecdata[j++] = rsblocks[row][col];
            }
          }
          /* and save out succesfully RS-decoded data */
          fp=fopen("RSdecdataC","wb");                       /* Output 256 bytes of user's data */
          fwrite(RSdecdata,1,sizeof(RSdecdata),fp);
          fclose(fp);
        }

        /* Print RS-decode status summary */
        if(Verbose){
          fprintf(stderr,"RS byte corrections: ");
          for(row=0;row<RSBLOCKS;row++){
            if(rserrs[row] != -1)
              fprintf(stderr," %d  ",rserrs[row]);
            else
              fprintf(stderr,"FAIL ");
          }
        }
      } /* end of rs section */


/* Step 4: Optional: Re-encode o/p and count channel errors  */
/* --------------------------------------------------------  */

     {  int errors = 0;
        int i;
        /* Input  array:  RSdecdata */
        /* Output array:  reencode  */

        encode_FEC40(reencode,RSdecdata) ;  /* Re-encode in AO-40 FEC format */

        /* Count the channel errors */
        errors = 0;
        for(i=0;i<SYMPBLOCK;i++)
          if ( reencode[i] != (raw[i]>>7) ) {
            errors++ ;
          }

        if(Verbose){
          fprintf(stderr,"Channel symbol errors: %d (%.3g%%)\n",errors,100.*errors/SYMPBLOCK);
        }

     }  /* end of reencode section */

// --------------------------------------------------------

        result := 0 ;
end { end of main } ;

end.

