(*
            FECRecon - ufrmFECRecon.pas
       (c) 2003 Paul C. L. Willmott, VP9MU

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created June 28, 2003 PCLW
     1.0.1      FEC decoding added PCLW
*)

unit ufrmFECRecon;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, StdCtrls, ExtCtrls, Menus, umdlAO40FECCodec ;

type

  TfrmFECRecon = class(TForm)
    btnRecon: TButton;
    lbledtInputPath: TLabeledEdit;
    lbledtOutputPath: TLabeledEdit;
    MainMenu1: TMainMenu;
    mnuFile: TMenuItem;
    mnuExit: TMenuItem;
    mnuHelp: TMenuItem;
    mnuAbout: TMenuItem;
    lbledtMovePath: TLabeledEdit;
    chkbxRaw: TCheckBox;
    procedure btnReconClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure mnuAboutClick(Sender: TObject);
    procedure mnuExitClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
    procedure Recon(sInFile,sOutFile:string) ;
    function Decode(sInFile,sOutFile:string):boolean ;
  end;

var
  frmFECRecon: TfrmFECRecon;

implementation

{$R *.dfm}

uses

     umdlOptions, ufrmAbout ;

{ TfrmFECRecon }

procedure TfrmFECRecon.Recon(sInFile, sOutFile: string);

var

     fInFile   : file of TRSDecData ;
     fOutFile  : file of TABlock ;
     pCodec    : TAO40FECCodec ;
     pABlock   : TABlock ;

begin

     pCodec := TAO40FECCodec.Create ;

     try

         AssignFile(fInFile,sInFile) ;
         Reset(fInFile) ;

         AssignFile(fOutFile,sOutFile) ;
         ReWrite(fOutFile) ;

         while (not Eof(fInFile)) do begin
             Read(fInFile,pCodec.pRSDecData) ;
             pCodec.RSfailures := false ;
             pABlock := pCodec.ReconABlock ;
             Write(fOutFile,pABlock) ;
         end ;

         CloseFile(fOutFile) ;
         CloseFile(fInFile) ;

     finally
          pCodec.Free ;
     end ;

end;

procedure TfrmFECRecon.btnReconClick(Sender: TObject);

var

     pSearchRec : TSearchRec ;
     sInFile, sOutFile, sMoveFile, sExt : string ;
     iSize : integer ;
     fFile : file of byte ;
     bIs256 : boolean ;
     bIs650 : boolean ;

procedure MoveFile(sSource,sDest:string) ;
var
     fIn, fOut : file ;
     pBuffer : array[0..65535] of byte ;
     iNumIn,iNumOut : integer ;
begin
     AssignFile(fIn,sSource) ;
     Reset(fIn,1) ;
     AssignFile(fOut,sDest) ;
     ReWrite(fOut,1) ;
     repeat
           BlockRead(fIn, pBuffer, SizeOf(pBuffer), iNumIn) ;
           BlockWrite(fOut, pBuffer, iNumIn, iNumOut) ;
     until (iNumIn = 0) or (iNumOut <> iNumIn) ;
     CloseFile(fOut) ;
     CloseFile(fIn) ;
     DeleteFile(sSource) ;
end { MoveFile } ;

begin // btnReconClick

     if (chkbxRaw.Checked) then begin
         sExt := '.RAW' ;
     end
     else begin
         sExt := '.TLM' ;
     end ;
     if (FindFirst(ExcludeTrailingBackslash(lbledtInputPath.Text)+'\*.FEC',0,pSearchRec)=0) then begin
         repeat
               sInFile   := ExcludeTrailingBackslash(lbledtInputPath.Text)  + '\' + pSearchRec.Name ;
               sOutFile  := ExcludeTrailingBackslash(lbledtOutputPath.Text) + '\' + ChangeFileExt(ExtractFileName(sInFile),sExt) ;
               sMoveFile := ExcludeTrailingBackslash(lbledtMovePath.Text)   + '\' + pSearchRec.Name ;
               AssignFile(fFile,sInFile) ;
               Reset(fFile) ;
               iSize := FileSize(fFile) ;
               CloseFile(fFile) ;
               bIs256 := iSize mod 256 = 0 ;
               bIs650 := iSize mod 650 = 0 ;
               if (not (bIs650 and Decode(sInFile,sOutFile))) and (bIs256) then begin
                   Recon(sInFile,sOutFile) ;    // ^ Short Circuit Boolean Evaluation Required For This To Work
               end ;
               MoveFile(sInFile,sMoveFile) ;
         until (FindNext(pSearchRec)<>0) ;
     end ;
     Application.MessageBox('All File(s) Reconstructed','FEC Block Reconstructor',MB_OK) ;
end;

procedure TfrmFECRecon.FormCreate(Sender: TObject);
begin
     Top := GetSetting(ksINIFile,Name,'Top',Top) ;
     Left := GetSetting(ksINIFile,Name,'Top',Left) ;
     lbledtInputPath.Text := GetSetting(ksINIFile,'Paths','Input Path','h:\AO40\Telemetry\To Be Processed\Decoded FEC') ;
     lbledtOutputPath.Text := GetSetting(ksINIFile,'Paths','Output Path','h:\AO40\Telemetry\To Be Processed') ;
     lbledtMovePath.Text := GetSetting(ksINIFile,'Paths','Move Path','h:\AO40\Telemetry\To Be Processed\Decoded FEC Done') ;
     chkbxRaw.Checked := GetSetting(ksINIFile,'Paths','Raw',chkbxRaw.Checked) ;
end;

procedure TfrmFECRecon.FormClose(Sender: TObject;
  var Action: TCloseAction);
begin
     SaveSetting(ksINIFile,Name,'Top',Top) ;
     SaveSetting(ksINIFile,Name,'Top',Left) ;
     SaveSetting(ksINIFile,'Paths','Input Path',lbledtInputPath.Text) ;
     SaveSetting(ksINIFile,'Paths','Output Path',lbledtOutputPath.Text) ;
     SaveSetting(ksINIFile,'Paths','Move Path',lbledtMovePath.Text) ;
     SaveSetting(ksINIFile,'Paths','Raw',chkbxRaw.Checked) ;
     Action := caFree ;
end;

procedure TfrmFECRecon.mnuAboutClick(Sender: TObject);
begin
     frmAbout := TfrmAbout.Create(Self) ;
     frmAbout.ShowModal ;
end;

procedure TfrmFECRecon.mnuExitClick(Sender: TObject);
begin
     Close ;
end;

function TfrmFECRecon.Decode(sInFile, sOutFile: string):boolean ;

var
     fInFile   : file of TRawBytes ;
     fOutFile  : file of TABlock ;
     pRawBytes : TRawBytes ;
     pCodec    : TAO40FECCodec ;
     pABlock   : TABlock ;

begin

     result := false ;
     pCodec := TAO40FECCodec.Create ;

     try

         AssignFile(fInFile,sInFile) ;
         Reset(fInFile) ;

         AssignFile(fOutFile,sOutFile) ;
         ReWrite(fOutFile) ;

         while (not Eof(fInFile)) do begin
             Read(fInFile,pRawBytes) ;
             pCodec.DecodeRawBytes(pRawBytes);
             if (not pCodec.RSfailures) or (chkbxRaw.Checked) then begin
                 pABlock := pCodec.ReconABlock ;
                 Write(fOutFile,pABlock) ;
                 result := true ;
             end ;
         end ;

         CloseFile(fOutFile) ;
         CloseFile(fInFile) ;

     finally
          pCodec.Free ;
     end ;

end;

end.
