(*
                  umdlAO40FECCodec
       (c) 2003 Paul C. L. Willmott, VP9MU

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Pascal Class Library Created July 12, 2003 PCLW

         based on:

         Reference Decoder for proposed AMSAT P3 Satellite Telemetry FEC Format
         Version 1.0 Last modified 2003 Jun 22 [Sun] 1035 utc (C)2003 J R Miller G3RUH
         This program is based on material largely developed by Phil Karn KA9Q over the
         period 1997-2002.   Much of it can be found in libraries on his www site at
         http://www.ka9q.net/ao40/  and includes notices which are reproduced herein.

         Other credits and copyrights appear within the text of this source file.

*)

unit umdlAO40FECCodec;

interface

const

// Defines for Interleaver

  kiROWS      = 80 ;                // Block interleaver rows
  kiCOLUMNS   = 65 ;                // Block interleaver columns
  kiSYMPBLOCK = kiROWS*kiCOLUMNS ;  // Encoded symbols per block

  kiBLOCKSIZE = 256 ; // Data bytes per frame
  kiRSBLOCKS  =   2 ; // Number of RS decoder blocks
  kiNN        = 255 ;
  kiKK        = 223 ;
  kiFCR       = 112 ;
  kiPRIM      =  11 ;
  kiIPRIM     = 116 ;

  kiA0        = kiNN ;
  kiRSPAD     = 95 ; // Unused bytes in block
  kiNROOTS    = 32 ;

  // Defines for Viterbi Decoder for r=1/2 k=7  (to CCSDS convention)

  kiK      = 7   ; // Constraint length
  kiN      = 2   ; // Number of symbols per data bit
  kiCPOLYA = $4f ; // First  convolutional encoder polynomial
  kiCPOLYB = $6d ; // Second convolutional encoder polynomial

// Number of symbols in an FEC block that are
// passed to the Viterbi decoder  (320*8 + 6)

  kiNBITS = (kiBLOCKSIZE+kiNROOTS*kiRSBLOCKS) * 8 + kiK -1 ;

  kpFECsync : array[0..64] of byte = (1,1,1,1,1,1,1,0,0,0,0,1,1,1,0,1,
                                      1,1,1,0,0,1,0,1,1,0,0,1,0,0,1,0,
                                      0,0,0,0,0,1,0,0,0,1,0,0,1,1,0,0,
                                      0,1,0,1,1,1,0,1,0,1,1,0,1,1,0,0,0) ;

type

  TRawBytes   = packed array[0..kiSYMPBLOCK div 8 -1] of byte ;
  TRawSymbols = packed array[0..kiSYMPBLOCK -1] of byte ;
  TSymbols    = packed array[0..5200] of byte ;
  TVitDecData = packed array[0..319] of byte ;
  TMetric     = packed array[0..63] of LongInt ;    // 2^(K-1)
  TRSDecData  = packed array[0..kiBLOCKSIZE-1] of byte ;
  TRSBlock    = packed array[0..kiNN -1] of byte ;

  TABlock = packed record
     case integer of
     1 : (
         pDataX : packed array[0..511] of byte ;
         ) ;
     2 : (
         pData1 : TRSDecData ;
         pData2 : TRSDecData ;
         ) ;
  end;

  TAO40FECCodec = class(TObject)
  private
    { Private declarations }
    pSymbols      : TSymbols ;
    pVitDecData   : TVitDecData ;
    FRSfailures: boolean;
    procedure SetRSfailures(const Value: boolean);
  public
    { Public declarations }
    pInputSymbols : TRawSymbols ;
    pRSDecData    : TRSDecData ;
    procedure DecodeRawBytes(var pRawBytes:TRawBytes) ;
    procedure DecodeRawSymbols(var pRawSymbols:TRawSymbols) ;
    procedure Viterbi27(var pSymbols:TSymbols;var pVitDecData:TVitDecData) ;
    function ReconABlock : TABlock ;
    function Decode_RS_8(var pRSblock:TRSBlock):integer ;
    property RSfailures : boolean  read FRSfailures write SetRSfailures;
  end ;

function HexStrW(wValue : word): string ;
function HexStrB(wValue : byte): string ;

implementation

uses

   SysUtils,
   Math ;

const

  kl1 : cardinal = 1 ;

(* Tables for Viterbi r=1/2 k=7 decoder to CCSDS standard */
/* ------------------------------------------------------ */

/* Metric table, [sent sym][rx symbol] */
/* This metric table is for an 8-bit ADC, which is total overkill!
   Simplify later.  128-i and i-128 would probably do!  jrm *)

kpMettab : array[0..1,0..255] of SmallInt =
((
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   19,   19,   19,   19,   19,   19,   19,   19,   19,
  19,   19,   18,   18,   18,   18,   18,   18,   17,   17,   17,   16,   16,   16,   15,   15,
  14,   14,   13,   13,   12,   11,   10,   10,    9,    8,    7,    6,    5,    3,    2,    1,
  -1,   -2,   -4,   -5,   -7,   -9,  -11,  -13,  -15,  -17,  -19,  -21,  -23,  -25,  -28,  -30,
 -32,  -35,  -37,  -40,  -42,  -45,  -47,  -50,  -52,  -55,  -58,  -60,  -63,  -66,  -68,  -71,
 -74,  -77,  -79,  -82,  -85,  -88,  -90,  -93,  -96,  -99, -102, -104, -107, -110, -113, -116,
-119, -121, -124, -127, -130, -133, -136, -138, -141, -144, -147, -150, -153, -155, -158, -161,
-164, -167, -170, -172, -175, -178, -181, -184, -187, -190, -192, -195, -198, -201, -204, -207,
-210, -212, -215, -218, -221, -224, -227, -229, -232, -235, -238, -241, -244, -247, -249, -252,
-255, -258, -261, -264, -267, -269, -272, -275, -278, -281, -284, -286, -289, -292, -295, -298,
-301, -304, -306, -309, -312, -315, -318, -320, -324, -326, -329, -332, -335, -337, -341, -372),
(
-372, -341, -338, -335, -332, -329, -326, -324, -321, -318, -315, -312, -309, -306, -304, -301,
-298, -295, -292, -289, -286, -284, -281, -278, -275, -272, -269, -267, -264, -261, -258, -255,
-252, -249, -247, -244, -241, -238, -235, -232, -229, -227, -224, -221, -218, -215, -212, -210,
-207, -204, -201, -198, -195, -192, -190, -187, -184, -181, -178, -175, -172, -170, -167, -164,
-161, -158, -155, -153, -150, -147, -144, -141, -138, -136, -133, -130, -127, -124, -121, -119,
-116, -113, -110, -107, -104, -102,  -99,  -96,  -93,  -90,  -88,  -85,  -82,  -79,  -77,  -74,
 -71,  -68,  -66,  -63,  -60,  -58,  -55,  -52,  -50,  -47,  -45,  -42,  -40,  -37,  -35,  -32,
 -30,  -28,  -25,  -23,  -21,  -19,  -17,  -15,  -13,  -11,   -9,   -7,   -5,   -4,   -2,   -1,
   1,    2,    3,    5,    6,    7,    8,    9,   10,   10,   11,   12,   13,   13,   14,   14,
  15,   15,   16,   16,   16,   17,   17,   17,   18,   18,   18,   18,   18,   18,   19,   19,
  19,   19,   19,   19,   19,   19,   19,   19,   19,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,
  20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20,   20
));


(* Table of symbol pair emitted for each state of the convolutional
 encoder's shift register *)

kpSyms : array[0..127] of byte = (
  1, 2, 3, 0, 2, 1, 0, 3, 2, 1, 0, 3, 1, 2, 3, 0,
  1, 2, 3, 0, 2, 1, 0, 3, 2, 1, 0, 3, 1, 2, 3, 0,
  0, 3, 2, 1, 3, 0, 1, 2, 3, 0, 1, 2, 0, 3, 2, 1,
  0, 3, 2, 1, 3, 0, 1, 2, 3, 0, 1, 2, 0, 3, 2, 1,
  2, 1, 0, 3, 1, 2, 3, 0, 1, 2, 3, 0, 2, 1, 0, 3,
  2, 1, 0, 3, 1, 2, 3, 0, 1, 2, 3, 0, 2, 1, 0, 3,
  3, 0, 1, 2, 0, 3, 2, 1, 0, 3, 2, 1, 3, 0, 1, 2,
  3, 0, 1, 2, 0, 3, 2, 1, 0, 3, 2, 1, 3, 0, 1, 2
);


// Scramble byte table

kpScrambler : array[0..319] of byte =
  (
   $ff, $48, $0e, $c0, $9a, $0d, $70, $bc, $8e, $2c, $93, $ad, $a7, $b7, $46, $ce,
   $5a, $97, $7d, $cc, $32, $a2, $bf, $3e, $0a, $10, $f1, $88, $94, $cd, $ea, $b1,
   $fe, $90, $1d, $81, $34, $1a, $e1, $79, $1c, $59, $27, $5b, $4f, $6e, $8d, $9c,
   $b5, $2e, $fb, $98, $65, $45, $7e, $7c, $14, $21, $e3, $11, $29, $9b, $d5, $63,
   $fd, $20, $3b, $02, $68, $35, $c2, $f2, $38, $b2, $4e, $b6, $9e, $dd, $1b, $39,
   $6a, $5d, $f7, $30, $ca, $8a, $fc, $f8, $28, $43, $c6, $22, $53, $37, $aa, $c7,
   $fa, $40, $76, $04, $d0, $6b, $85, $e4, $71, $64, $9d, $6d, $3d, $ba, $36, $72,
   $d4, $bb, $ee, $61, $95, $15, $f9, $f0, $50, $87, $8c, $44, $a6, $6f, $55, $8f,
   $f4, $80, $ec, $09, $a0, $d7, $0b, $c8, $e2, $c9, $3a, $da, $7b, $74, $6c, $e5,
   $a9, $77, $dc, $c3, $2a, $2b, $f3, $e0, $a1, $0f, $18, $89, $4c, $de, $ab, $1f,
   $e9, $01, $d8, $13, $41, $ae, $17, $91, $c5, $92, $75, $b4, $f6, $e8, $d9, $cb,
   $52, $ef, $b9, $86, $54, $57, $e7, $c1, $42, $1e, $31, $12, $99, $bd, $56, $3f,
   $d2, $03, $b0, $26, $83, $5c, $2f, $23, $8b, $24, $eb, $69, $ed, $d1, $b3, $96,
   $a5, $df, $73, $0c, $a8, $af, $cf, $82, $84, $3c, $62, $25, $33, $7a, $ac, $7f,
   $a4, $07, $60, $4d, $06, $b8, $5e, $47, $16, $49, $d6, $d3, $db, $a3, $67, $2d,
   $4b, $be, $e6, $19, $51, $5f, $9f, $05, $08, $78, $c4, $4a, $66, $f5, $58, $ff,
   $48, $0e, $c0, $9a, $0d, $70, $bc, $8e, $2c, $93, $ad, $a7, $b7, $46, $ce, $5a,
   $97, $7d, $cc, $32, $a2, $bf, $3e, $0a, $10, $f1, $88, $94, $cd, $ea, $b1, $fe,
   $90, $1d, $81, $34, $1a, $e1, $79, $1c, $59, $27, $5b, $4f, $6e, $8d, $9c, $b5,
   $2e, $fb, $98, $65, $45, $7e, $7c, $14, $21, $e3, $11, $29, $9b, $d5, $63, $fd
) ;



(* Tables for RS decoder */
/* Galois field log/antilog tables *)

kpALPHA_TO : array[0..255] of integer =
  (
    $01, $02, $04, $08, $10, $20, $40, $80, $87, $89, $95, $ad, $dd, $3d, $7a, $f4,
    $6f, $de, $3b, $76, $ec, $5f, $be, $fb, $71, $e2, $43, $86, $8b, $91, $a5, $cd,
    $1d, $3a, $74, $e8, $57, $ae, $db, $31, $62, $c4, $0f, $1e, $3c, $78, $f0, $67,
    $ce, $1b, $36, $6c, $d8, $37, $6e, $dc, $3f, $7e, $fc, $7f, $fe, $7b, $f6, $6b,
    $d6, $2b, $56, $ac, $df, $39, $72, $e4, $4f, $9e, $bb, $f1, $65, $ca, $13, $26,
    $4c, $98, $b7, $e9, $55, $aa, $d3, $21, $42, $84, $8f, $99, $b5, $ed, $5d, $ba,
    $f3, $61, $c2, $03, $06, $0c, $18, $30, $60, $c0, $07, $0e, $1c, $38, $70, $e0,
    $47, $8e, $9b, $b1, $e5, $4d, $9a, $b3, $e1, $45, $8a, $93, $a1, $c5, $0d, $1a,
    $34, $68, $d0, $27, $4e, $9c, $bf, $f9, $75, $ea, $53, $a6, $cb, $11, $22, $44,
    $88, $97, $a9, $d5, $2d, $5a, $b4, $ef, $59, $b2, $e3, $41, $82, $83, $81, $85,
    $8d, $9d, $bd, $fd, $7d, $fa, $73, $e6, $4b, $96, $ab, $d1, $25, $4a, $94, $af,
    $d9, $35, $6a, $d4, $2f, $5e, $bc, $ff, $79, $f2, $63, $c6, $0b, $16, $2c, $58,
    $b0, $e7, $49, $92, $a3, $c1, $05, $0a, $14, $28, $50, $a0, $c7, $09, $12, $24,
    $48, $90, $a7, $c9, $15, $2a, $54, $a8, $d7, $29, $52, $a4, $cf, $19, $32, $64,
    $c8, $17, $2e, $5c, $b8, $f7, $69, $d2, $23, $46, $8c, $9f, $b9, $f5, $6d, $da,
    $33, $66, $cc, $1f, $3e, $7c, $f8, $77, $ee, $5b, $b6, $eb, $51, $a2, $c3, $00
 ) ;

kpINDEX_OF : array[0..255] of integer =
  (
   $ff, $00, $01, $63, $02, $c6, $64, $6a, $03, $cd, $c7, $bc, $65, $7e, $6b, $2a,
   $04, $8d, $ce, $4e, $c8, $d4, $bd, $e1, $66, $dd, $7f, $31, $6c, $20, $2b, $f3,
   $05, $57, $8e, $e8, $cf, $ac, $4f, $83, $c9, $d9, $d5, $41, $be, $94, $e2, $b4,
   $67, $27, $de, $f0, $80, $b1, $32, $35, $6d, $45, $21, $12, $2c, $0d, $f4, $38,
   $06, $9b, $58, $1a, $8f, $79, $e9, $70, $d0, $c2, $ad, $a8, $50, $75, $84, $48,
   $ca, $fc, $da, $8a, $d6, $54, $42, $24, $bf, $98, $95, $f9, $e3, $5e, $b5, $15,
   $68, $61, $28, $ba, $df, $4c, $f1, $2f, $81, $e6, $b2, $3f, $33, $ee, $36, $10,
   $6e, $18, $46, $a6, $22, $88, $13, $f7, $2d, $b8, $0e, $3d, $f5, $a4, $39, $3b,
   $07, $9e, $9c, $9d, $59, $9f, $1b, $08, $90, $09, $7a, $1c, $ea, $a0, $71, $5a,
   $d1, $1d, $c3, $7b, $ae, $0a, $a9, $91, $51, $5b, $76, $72, $85, $a1, $49, $eb,
   $cb, $7c, $fd, $c4, $db, $1e, $8b, $d2, $d7, $92, $55, $aa, $43, $0b, $25, $af,
   $c0, $73, $99, $77, $96, $5c, $fa, $52, $e4, $ec, $5f, $4a, $b6, $a2, $16, $86,
   $69, $c5, $62, $fe, $29, $7d, $bb, $cc, $e0, $d3, $4d, $8c, $f2, $1f, $30, $dc,
   $82, $ab, $e7, $56, $b3, $93, $40, $d8, $34, $b0, $ef, $26, $37, $0c, $11, $44,
   $6f, $78, $19, $9a, $47, $74, $a7, $c1, $23, $53, $89, $fb, $14, $5d, $f8, $97,
   $2e, $4b, $b9, $60, $0f, $ed, $3e, $e5, $f6, $87, $a5, $17, $3a, $a3, $3c, $b7

  ) ;


function HexStrW(wValue: word): string;
const
     _Hex : string = '0123456789ABCDEF' ;
begin
     result := _Hex[1 + (wValue and $f000) shr 12] +
               _Hex[1 + (wValue and $0f00) shr  8] +
               _Hex[1 + (wValue and $00f0) shr  4] +
               _Hex[1 + (wValue and $000f)       ] ;
end;

function HexStrB(wValue: byte): string;
const
     _Hex : string = '0123456789ABCDEF' ;
begin
     result := _Hex[1 + (wValue and $00f0) shr  4] +
               _Hex[1 + (wValue and $000f)       ] ;
end;

{ TAO40FECCodec }

// Adapted from the work of VK5HI.

procedure TAO40FECCodec.DecodeRawBytes(var pRawBytes: TRawBytes);

var
     iIndex, iBit, iOutIndex : integer ;

begin

//   Create "fudged" symbols by VK5HI Method

     iOutIndex := 0 ;
     for iIndex := 0 to kiSYMPBLOCK div 8 -1 do begin
         for iBit := 0 to 7 do begin
             if ((pRawBytes[iIndex] shl iBit) and $80 = $80) then begin
                 pInputSymbols[iOutIndex] := $f0 ;
             end
             else begin
                 pInputSymbols[iOutIndex] := $10 ;
             end ;
             Inc(iOutIndex) ;
         end ;
     end ;

//   Continue with regular G3RUH method

     DecodeRawSymbols(pInputSymbols) ;

end;

(****************************************************************************************************/
/*                                                                                                  */
/*   Based upon the C code of G3RUH                                                                 */
/*                                                                                                  */
/*   AMSAT AO-40 FEC Reference Decoder (by JRM)               Input                Output           */
/*   ------------------------------------------               -----------------------------------   */
/*   Step 1: De-interleave                                    pRawSymbols[5200] ->  pSymbols[5132]  */
/*   Step 2: Viterbi decoder                                  pSymbols[5132]  ->  pVitDecData[320]  */
/*   Step 3: RS decoder                                       pVitDecData[320] -> pRSdecdata[256]   */
/*   Step 4: Option: Re-encode o/p and count channel errors   pRSdecdata[256] ->  reencode[5200]    */
/*                                                                                                  */
/****************************************************************************************************)

procedure TAO40FECCodec.DecodeRawSymbols(var pRawSymbols: TRawSymbols);

var

   iCol, iRow, iColTop, iRowStart : integer ;

   pRSblocks : array [0..kiRSBLOCKS-1] of TRSBlock ;
   pRSerrs   : array [0..kiRSBLOCKS-1] of integer ;  // Count of errors in each RS codeword
   iIndex    : integer ;

begin

// Step 1: De-interleaver
// ----------------------

     iColTop := 0 ;
     for iCol := 1 to kiROWS -1 do begin // Skip first column as it's the sync vector
         iRowStart := 0 ;
         for iRow :=0 to kiCOLUMNS -1 do begin
             pSymbols[iColTop+iRow] := pRawSymbols[iRowStart+iCol] ;  // coltop=col*65 ; rowstart=row*80
             Inc(iRowStart,kiROWS) ;
         end ;
         Inc(iColTop,kiCOLUMNS) ;
     end ;

// Step 2: Viterbi decoder
// -----------------------

     Viterbi27(pSymbols,pVitDecData) ; // Parameter order switched to Input, Output

// Step 3: RS decoder
// ------------------

(* There are two RS decoders, processing 128 bytes each.
 *
 * If both RS decoders are SUCCESSFUL
 * On exit:
 *   rs_failures = 0
 *   rserrs[x]   = number of errors corrected;  range 0 to 16   (x= 0 or 1)
 *   Data output is in array RSdecdata[256].
 *
 * If an RS decoder FAILS
 * On exit:
 *   rs_failures = 1 or 2   (i.e. != 0)
 *   rserrs[x] contains -1
 *   Data output should not be used.
 *)

//   Interleave into Reed Solomon codeblocks

     FillChar(pRSBlocks,SizeOf(pRSBlocks),0); // Zero pRSblocks array

//   Remove scrambling

     iIndex := 0 ;
     for iCol := kiRSPAD to kiNN -1 do begin
         for iRow := 0 to kiRSBLOCKS -1 do begin
             pRSblocks[iRow,iCol] := pVitDecData[iIndex] xor kpScrambler[iIndex] ;
             Inc(iIndex) ;
         end ;
     end ;

//   Run RS-decoder(s)

     FRSfailures := false ;

     for iRow := 0 to kiRSBLOCKS-1 do begin
         pRSerrs[iRow] := Decode_RS_8(pRSblocks[iRow]) ;
         FRSfailures := FRSfailures or (pRSerrs[iRow]=-1) ;
     end ;

//   If frame decoded OK, deinterleave data from RS codeword(s)

     if (not FRSfailures) then begin
         iIndex := 0 ;
         for iCol := kiRSPAD to kiKK -1 do begin
             for iRow := 0 to kiRSBLOCKS -1 do begin
    	         pRSdecdata[iIndex] := pRSblocks[iRow,iCol] ;
                 Inc(iIndex) ;
             end ;
         end ;
     end ;

end;

function TAO40FECCodec.Decode_RS_8(var pRSblock: TRSBlock): integer;

(* This decoder has evolved extensively through the work of Phil Karn.  It draws
 * on his own ideas and optimisations, and on the work of others.  The lineage
 * is as below, and parts of the authors' notices are included here.  (JRM)

 * Reed-Solomon decoder
 * Copyright 2002 Phil Karn, KA9Q
 * May be used under the terms of the GNU General Public License (GPL)
 *
 * Reed-Solomon coding and decoding
 * Phil Karn (karn@ka9q.ampr.org) September 1996
 *
 * This file is derived from the program "new_rs_erasures.c" by Robert
 * Morelos-Zaragoza (robert@spectra.eng.hawaii.edu) and Hari Thirumoorthy
 * (harit@spectra.eng.hawaii.edu), Aug 1995
 * --------------------------------------------------------------------------
 *
 * From the RM-Z & HT program:
 * The encoding and decoding methods are based on the
 * book "Error Control Coding: Fundamentals and Applications",
 * by Lin and Costello, Prentice Hall, 1983, ISBN 0-13-283796-X
 * Portions of this program are from a Reed-Solomon encoder/decoder
 * in C, written by Simon Rockliff (simon@augean.ua.oz.au) on 21/9/89.
 * --------------------------------------------------------------------------
 *
 * From the 1989/1991 SR program (also based on Lin and Costello):
 * This program may be freely modified and/or given to whoever wants it.
 * A condition of such distribution is that the author's contribution be
 * acknowledged by his name being left in the comments heading the program,
 *                               Simon Rockliff, 26th June 1991
 *
 *)

label

     finish,
     label1 ; // label1 used to fake "C" continue

type

     TRoot = array [0..kiNROOTS] of byte ;

     var

     deg_lambda, el, deg_omega : integer ;
     i, j, r, k : integer ;
     u,q,tmp,num1,num2,den,discr_r : byte ;
     syn_error, count : integer ;

     mp : integer ; // Move Pointer

     lambda ,
     s      ,   // Err+Eras Locator poly and syndrome poly
     b      ,
     t      ,
     omega  ,
     root   ,
     reg    ,
     loc    : TRoot ;

function mod255(x:integer):integer ;
begin
     while (x >= 255) do begin
         Dec(x,255) ;
         x := (x shr 8) + (x and 255);
     end ;
     result := x ;
end { mod255 } ;

begin

// form the syndromes; i.e., evaluate data(x) at roots of g(x)

     for i := 0 to kiNROOTS-1 do begin
         s[i] := pRSblock[0] ;
     end ;

     for j := 1 to kiNN-1 do begin
         for i := 0 to kiNROOTS-1 do begin
             if (s[i] = 0) then begin
                 s[i] := pRSblock[j] ;
             end
             else begin
                 s[i] := pRSblock[j] xor kpALPHA_TO[mod255(kpINDEX_OF[s[i]] + (kiFCR+i)*kiPRIM)] ;
             end ;
          end ;
     end ;

// Convert syndromes to index form, checking for nonzero condition */

     syn_error := 0 ;
     for i := 0 to kiNROOTS-1 do begin
         syn_error := syn_error or s[i] ;
         s[i] := kpINDEX_OF[s[i]] ;
     end ;

//   if syndrome is zero, data[] is a codeword and there are no
//   errors to correct. So return data[] unmodified

     if (syn_error=0) then begin
         count := 0 ;
         goto finish ;
     end ;

     FillChar(lambda,SizeOf(lambda),0) ;
     lambda[0] := 1 ;

     for i := 0 to kiNROOTS do begin
         b[i] := kpINDEX_OF[lambda[i]] ;
     end ;

//   Begin Berlekamp-Massey algorithm to determine error+erasure
//   locator polynomial

     el := 0 ;
     r  := 1 ;
     while (r <= kiNROOTS) do begin       // r is the step number

//   Compute discrepancy at the r-th step in poly-form

         discr_r := 0 ;
         for i := 0 to r-1 do begin
              if ((lambda[i] <> 0) and (s[r-i-1] <> kiA0)) then begin
                  discr_r := discr_r xor kpALPHA_TO[mod255(kpINDEX_OF[lambda[i]] + s[r-i-1])] ;
              end ;
         end ;

         discr_r := kpINDEX_OF[discr_r] ;        // Index form

         if (discr_r = kiA0) then begin

//   2 lines below: B(x) <-- x*B(x)

             for mp := kiNROOTS downto 1 do begin
                 b[mp] := b[mp-1] ;
             end ;
             b[0] := kiA0 ;

         end
         else begin

//   7 lines below: T(x) <-- lambda(x) - discr_r*x*b(x)

             t[0] := lambda[0] ;
             for i := 0 to kiNROOTS -1 do begin
                  if (b[i] <> kiA0) then begin
                     t[i+1] := lambda[i+1] xor kpALPHA_TO[mod255(discr_r + b[i])] ;
                  end
                  else begin
                     t[i+1] := lambda[i+1] ;
                  end ;
             end ;

             if (2 * el <= r - 1) then begin

                 el := r - el ;

//   2 lines below: B(x) <-- inv(discr_r) * lambda(x)

                 for i := 0 to kiNROOTS do begin
                     if (lambda[i] = 0) then begin
                         b[i] := kiA0 ;
                     end
                     else begin
                         b[i] := mod255(kpINDEX_OF[lambda[i]] - discr_r + kiNN) ;
                     end ;
                 end ;

             end
             else begin

//   2 lines below: B(x) <-- x*B(x)

                 for mp := kiNROOTS downto 1 do begin
                     b[mp] := b[mp-1] ;
                 end ;
                 b[0] := kiA0 ;

             end ;

             lambda := t ;

         end ;

         Inc(r) ;

     end ;

//   Convert lambda to index form and compute deg(lambda(x))

     deg_lambda := 0 ;
     for i :=0 to kiNROOTS do begin
         lambda[i] := kpINDEX_OF[lambda[i]] ;
         if (lambda[i] <> kiA0) then begin
             deg_lambda := i ;
         end ;
     end ;

//   Find roots of the error+erasure locator polynomial by Chien search

     for mp := 1 to kiNROOTS do begin
         reg[mp] := lambda[mp] ;
     end ;

     count := 0 ;            // Number of roots of lambda(x)

     i := 1 ;
     k := kiIPRIM-1 ;
     while (i <= kiNN) do begin

         q := 1 ; // lambda[0] is always 0

         for j := deg_lambda downto 1 do begin
             if (reg[j] <> kiA0) then begin
                 reg[j] := mod255(reg[j] + j) ;
                 q := q xor kpALPHA_TO[reg[j]];
             end ;
         end ;

         if (q <> 0) then begin
             goto label1 ; // Not a root
         end ;

//   store root (index-form) and error location number

         root[count] := i ;
         loc[count] := k ;

//   If we've already found max possible roots,
//   abort the search to save time

         Inc(count) ;
         if (count = deg_lambda) then begin
             break ;
         end ;

label1: // Pascal Continue is different to C, as I found out after 8 hours debugging!

         Inc(i) ;
         k := mod255(k+kiIPRIM) ;

     end ;

//   deg(lambda) unequal to number of roots => uncorrectable
//   error detected

     if (deg_lambda <> count) then begin
         count := -1 ;
         goto finish ;
     end ;

//   Compute err+eras evaluator poly omega(x) = s(x)*lambda(x) (modulo
//   x**NROOTS). in index form. Also find deg(omega).

     deg_omega := 0 ;
     for i := 0 to kiNROOTS -1 do begin
          tmp := 0 ;
          if (deg_lambda < i) then begin
              j := deg_lambda ;
          end
          else begin
              j := i ;
          end ;
          while j >= 0 do begin
              if ((s[i - j] <> kiA0) and (lambda[j] <> kiA0)) then begin
                  tmp := tmp xor kpALPHA_TO[mod255(s[i - j] + lambda[j])] ;
              end ;
              Dec(j) ;
          end ;
          if(tmp <> 0) then begin
             deg_omega := i ;
          end ;
          omega[i] := kpINDEX_OF[tmp] ;
     end ;
     omega[kiNROOTS] := kiA0 ;

//   Compute error values in poly-form. num1 = omega(inv(X(l))), num2 =
//   inv(X(l))**(FCR-1) and den = lambda_pr(inv(X(l))) all in poly-form

     for j := count-1 downto 0 do begin

         num1 := 0 ;
         for i := deg_omega downto 0 do begin
              if (omega[i] <> kiA0) then begin
                  num1 := num1 xor (kpALPHA_TO[mod255(omega[i] + i * root[j])]) ;
              end ;
         end ;
         num2 := kpALPHA_TO[mod255(root[j] * (kiFCR - 1) + kiNN)] ;
         den := 0 ;

//   lambda[i+1] for i even is the formal derivative lambda_pr of lambda[i]

         i := Min(deg_lambda,kiNROOTS-1) and not 1 ;
         while (i >= 0) do begin
             if (lambda[i+1] <> kiA0) then begin
                den := den xor (kpALPHA_TO[mod255(lambda[i+1] + i * root[j])]);
             end ;
             Dec(i,2) ;
         end ;

         if (den = 0) then begin
             count := -1 ;
             goto finish ;
         end ;

//   Apply error to data

         if (num1 <> 0) then begin
             pRSblock[loc[j]] := pRSblock[loc[j]] xor (kpALPHA_TO[mod255(kpINDEX_OF[num1] + kpINDEX_OF[num2] + kiNN - kpINDEX_OF[den])]) ;
         end ;

     end ;

finish:

     result := count ;

end { Decode_RS_8 } ;

function TAO40FECCodec.ReconABlock: TABlock;
var
     pBlock : TABlock ;
     dDate  : tDateTime ;

procedure WriteStr(wAddr: word; sStr: string);
var
     iIndex : integer ;
begin
     for iIndex := 1 to Length(sStr) do begin
         pBlock.pDataX[wAddr+pred(iIndex)] := BYTE(sStr[iIndex]) ;
     end ;
end;

begin // ReconABlock

//                           1         2         3         4         5         6
//                 0123456789012345678901234567890123456789012345678901234567890123
     WriteStr(  0,'A  HI, THIS IS AMSAT OSCAR-40       yyyy-mm-dd  hh:mm:ss  #nnnn ') ;
     WriteStr( 64,'+--------------------------------------------------------------+') ;
     WriteStr(128,'|             RECONSTRUCTED FROM DECODED FEC BLOCK             |') ;
     WriteStr(192,'+--------------------------------------------------------------+') ;

     pBlock.pData2 := pRSDecData ;

     dDate := EncodeDate(1978,1,1)                               +
              pBlock.pDataX[$1AD] * 256.0                        +
              pBlock.pDataX[$1AC]                                +
              pBlock.pDataX[$1AB] / (24.0                      ) +
              pBlock.pDataX[$1AA] / (24.0 * 60.0               ) +
              pBlock.pDataX[$1A9] / (24.0 * 60.0 * 60.0        ) +
              pBlock.pDataX[$1A8] / (24.0 * 60.0 * 60.0 * 100.0) ;

     WriteStr(36,FormatDateTime('yyyy-mm-dd  hh:nn:ss',dDate)) ;

     WriteStr(58,'#' + HexStrW(pBlock.pDataX[$1E1] * 256 + pBlock.pDataX[$1E0])) ;

     if (RSFailures) then begin
         pBlock.pDataX[0] := pBlock.pDataX[0] or $80 ;
     end ;

     result := pBlock ;

end;

procedure TAO40FECCodec.SetRSfailures(const Value: boolean);
begin
  FRSfailures := Value;
end;

procedure TAO40FECCodec.Viterbi27(var pSymbols: TSymbols;
  var pVitDecData: TVitDecData);

(* Viterbi decoder for arbitrary convolutional code
 * viterbi27 and viterbi37 for the r=1/2 and r=1/3 K=7 codes are faster
 * Copyright 1997 Phil Karn, KA9Q
 *
 * This is a bare bones <2,7> Viterbi decoder, adapted from a general purpose model.
 * It is not optimised in any way, neither as to coding (for example the memcopy should
 * be achievable simply by swapping pointers), nor as to simplifying the metric table,
 * nor as to using any machine-specific smarts.  On contemporary machines, in this application,
 * the execution time is negligible.  Many ideas for optimisation are contained in PK's www pages.
 * The real ADC is 8-bit, though in practice 4-bits are actually sufficient.
 * Descriptions of the Viterbi decoder algorithm can be found in virtually any book
 * entitled "Digital Communications".  (JRM)
 *)

var

     pCmetric   : TMetric ;
     pNmetric   : TMetric ;
     pMets      : array[0.. 3] of Integer ;    // 2^N
     iIndex1    : Integer ;
     iIndex2    : Integer ;
     iSymbol    : Integer ; // Can't use C trick

     pPaths     : array[0..kiNBITS*2-1] of cardinal ;
     iPathPtr   : Integer ; // Can't use C trick

     iBitCnt    : Cardinal ;
     iBestState : Integer ;

     iB1, iB2 : Integer ;
     iM0, iM1, iMask : Integer ;

begin

//   Initialize starting metrics to prefer 0 state

     pCmetric[0] := 0 ;
     for iIndex1 := 1 to 63 do begin
         pCmetric[iIndex1] := -999999 ;
     end ;

     FillChar(pPaths,SizeOf(pPaths),0) ;

     iBitCnt  := 0 ;
     iPathPtr := 0 ;
     iSymbol  := 0 ;

     repeat

//   Read 2 input symbols and compute the 4 branch metrics

           for iIndex1 := 0 to 3 do begin
               pMets[iIndex1] := 0 ;
               for iIndex2 := 0 to 1 do begin
                   pMets[iIndex1] := pMets[iIndex1] + kpMettab[(iIndex1 shr (1-iIndex2)) and 1, pSymbols[iSymbol+iIndex2]] ;
               end ;
           end ;
           Inc(iSymbol,2) ;

           iMask := 1 ;
           iIndex1 := 0 ;
           while (iIndex1<64) do begin
               iB1 := pMets[kpSyms[iIndex1]] ;
               iM0 := pCmetric[iIndex1 div 2] + iB1 ;
               pNmetric[iIndex1] := iM0 ;
               iB2 := pMets[kpSyms[iIndex1+1]] ;
               Dec(iB1,iB2)  ;
               iM1 := pCmetric[(iIndex1 div 2) + (1 shl (kiK-2))] + iB2 ;
               if (iM1 > iM0) then begin
                  pNmetric[iIndex1] := iM1 ;
                  pPaths[iPathPtr] := pPaths[iPathPtr] or iMask ;
               end ;
               Dec(iM0,iB1) ;
               pNmetric[iIndex1+1] := iM0 ;
               Inc(iM1,iB1) ;
               if (iM1 > iM0) then begin
                   pNmetric[iIndex1+1] := iM1 ;
                   pPaths[iPathPtr] := pPaths[iPathPtr] or (cardinal(iMask) shl 1) ;
               end ;
               iMask := iMask shl 2 ;
               if (iMask = 0) then begin
                   iMask := 1 ;
                   iPathPtr := iPathPtr + 1 ;
               end ;
               Inc(iIndex1,2) ;
           end ;
           if (iMask <> 1) then begin
               iPathPtr := iPathPtr + 1 ;
           end ;
           Inc(iBitCnt) ;
           if (iBitCnt = kiNBITS) then begin
               iBestState := 0 ;
               break ;
           end ;
           pCMetric := pNmetric ;
     until false ;

     Dec(iPathPtr,2) ;

//   Chain back from terminal state to produce decoded data

     Fillchar(pVitDecData,SizeOf(pVitDecData),0) ;

     iIndex1 := kiNBITS - kiK ;
     while (iIndex1>=0) do begin
         if (pPaths[iPathPtr + (iBestState shr 5)] and (1 shl (iBestState and 31))<>0) then begin
             iBestState := iBestState or (1 shl (kiK-1)) ;
             pVitDecData[iIndex1 shr 3] := pVitDecData[iIndex1 shr 3] or ($80 shr (iIndex1 and 7)) ;
         end ;
         iBestState := iBestState shr 1 ;
         Dec(iPathPtr,2) ;
         Dec(iIndex1) ;
     end ;

end { Viterbi27 } ;

end.
