(*
                  umdlOptions
       (c) 2003 Paul C. L. Willmott, VP9MU

 This program is free software; you can redistribute it
 and/or modify it under the terms of the GNU General
 Public License as published by the Free Software
 Foundation; either version 2 of the License, or at
 your option, any later version.

 This program is distributed in the hope that it will
 be useful, but WITHOUT ANY WARRANTY; without even the
 implied warranty of MERCHANTABILITY or FITNESS FOR A
 PARTICULAR PURPOSE.  See the GNU General Public
 License for more details.

 You should have received a copy of the GNU General
 Public License along with this program; if not, write
 to the Free Software Foundation, Inc., 59 Temple
 Place, Suite 330, Boston, MA  02111-1307  USA

            Contact : vp9mu@amsat.org
*)

(*
     Revision History
     ----------------

     1.0.0      Created June 1, 2003 PCLW
     1.0.1      June 16, 2003 GetSetting & SaveSetting added PCLW

*)

unit umdlOptions;

interface

const

      ksINIFile : string = 'FECRecon.ini' ;

function GetSetting(sFile,sSection,sKey,Default:string):string ; overload ;
function GetSetting(sFile,sSection,sKey:string;Default:integer):integer ; overload ;
function GetSetting(sFile,sSection,sKey:string;Default:boolean):boolean ; overload ;
procedure SaveSetting(sFile,sSection,sKey,Value:string) ; overload ;
procedure SaveSetting(sFile,sSection,sKey:string;Value:integer) ; overload ;
procedure SaveSetting(sFile,sSection,sKey:string;Value:boolean) ; overload ;

implementation

uses

     INIFiles ;

function GetSetting(sFile,sSection,sKey,Default:string):string ;
var
     iIndex   : integer  ;
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         result := pINIFile.ReadString(sSection,sKey,Default) ;
     finally
         pINIFile.Free ;
     end ;
end  { GetSetting } ;

function GetSetting(sFile,sSection,sKey:string;Default:integer):integer ;
var
     iIndex   : integer  ;
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         result := pINIFile.ReadInteger(sSection,sKey,Default) ;
     finally
         pINIFile.Free ;
     end ;
end  { GetSetting } ;

function GetSetting(sFile,sSection,sKey:string;Default:boolean):boolean ;
var
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         result := pINIFile.ReadBool(sSection,sKey,Default) ;
     finally
         pINIFile.Free ;
     end ;
end  { GetSetting } ;

procedure SaveSetting(sFile,sSection,sKey,Value:string) ;
var
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         pINIFile.WriteString(sSection,sKey,Value);
         pINIFile.UpdateFile ; // Flush memory copy to disc
     finally
         pINIFile.Free ;
     end ;
end { SaveSetting } ;

procedure SaveSetting(sFile,sSection,sKey:string;Value:integer) ;
var
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         pINIFile.WriteInteger(sSection,sKey,Value);
         pINIFile.UpdateFile ; // Flush memory copy to disc
     finally
         pINIFile.Free ;
     end ;
end { SaveSetting } ;

procedure SaveSetting(sFile,sSection,sKey:string;Value:boolean) ;
var
     pINIFile : TINIFile ;
begin
     pINIFile := TINIFile.Create(sFile) ;
     try
         pINIFile.WriteBool(sSection,sKey,Value);
         pINIFile.UpdateFile ; // Flush memory copy to disc
     finally
         pINIFile.Free ;
     end ;
end { SaveSetting } ;

end.